# -*- coding: utf-8 -*-
# vi:ts=4:et
#
# $Date: 2005/04/16 07:18:45 $
# $Revision: 1.13 $
# =====================================================

"""library for permutation and combination"""

import operator
import itertools

import util

__all__ = [
           'factorial',
           'permutation',
           'combination',
           'catalan',
           'catalan_generator',
           'k_composition',
           'weak_k_composition',
           'signless_stirling1',
           'stirling1',
           'stirling2',
           ]

def factorial(x, y=0):
    """factorial([m=1), n) -> Return the factorial of n.

    If m is given, multiply an iterable from m to n.
    i.e., m x (m+1) x ... x n

    >>> for i in range(10):
    ...     print factorial(i)
    ...
    1
    1
    2
    6
    24
    120
    720
    5040
    40320
    362880
    """
    if y:
        start, stop = x, y
    else:
        start, stop = 2, x

    # factorial(2,5) -> 2*3*4*5
    # factorial(3,4) -> 3*4

    # factorial(4) = 4! -> 4*3*2*1 = 24
    # factorial(3) = 3! -> 3*2*1 = 6

    assert isinstance(stop, (int, long)) and stop >= 0, "argument must be a non-negative integer."

    assert isinstance(start, (int, long)) and start >= 0, "argument must be a non-negative integer."


    if stop < 2:
        return 1

    assert start <= stop, \
            "first argument <= second argument is needed." +\
            "%s > %s is given."%(start, stop)

    return _mul_of_iterable(xrange(start, stop + 1), 1)

def _mul_of_iterable(iterable, initial_value=0):
    """Return the multiply of iterable.
    """
    # If iterable is [2,4,6]
    # _mul_of_iterable(iterable) = 2 * 4 * 6 = 48

    if initial_value:
        return reduce(operator.mul, iterable, initial_value)
    else:
        return reduce(operator.mul, iterable)

def combination(first, second):
    """combination(n, m) -> Return the combination of (n,m).
    n can be a sequence or an integer; m must be an integer.

    If n is an integer, return the combination of (n, m)
    If n is a sequence, take m elements out of the sequence.

    >>> n = 5
    >>> for i in range(n):
    ...     print combination(n, i)
    ...
    1
    5
    10
    10
    5

    >>> s = 'abcd'
    >>> for i in range(len(s) + 1):
    ...     print combination(s, i)
    ...
    [[]]
    [['a'], ['b'], ['c'], ['d']]
    [['a', 'b'], ['a', 'c'], ['a', 'd'], ['b', 'c'], ['b', 'd'], ['c', 'd']]
    [['a', 'b', 'c'], ['a', 'b', 'd'], ['a', 'c', 'd'], ['b', 'c', 'd']]
    [['a', 'b', 'c', 'd']]

    """
    assert isinstance(second, (int, long)) and second >= 0, "second argument must be a non-negative integer."

    if isinstance(first, (int, long)):
        if second == 0:
            return 1
        else:
            return _combination_of_integer(first, second)

    # Now we know that the first argument is not an integer,
    # the first needs to be a sequence.

    try:
        len(first)
    except TypeError, e:
        raise TypeError, "bad operand type was found in the argument : %s"%first
    else:
        return _combination_of_sequence(first, second)

def _combination_of_sequence(seq, num):
    """out of the sequence, take num elements
    """
    if len(seq) < num:
        num = len(seq)
    #assert len(seq) >= num,  seq

    return list(com_gen(seq, num))

def _combination_of_integer(big, small):
    """combination of two integers.
    """
    # (4,2)  -> 4*3 / 2*1 = 6
    # (n, m) -> n!/(m! * (n-m)!)
    # = n*(n-1) * ... * (n-m+1)/m!
    assert big >= small >=0,\
            "invalid argument (%s, %s)\n"%(big, small) + \
            " "*4 +\
            "first argument must be greater than or equal to second argument."

    return factorial(big - small +1, big)/factorial(small)

def permutation(*seq):
    """permutation(sequence) -> Return the permutation of sequence

    >>> seq = range(3)
    >>> for x in sorted(permutation(seq)):
    ...     print x
    ...
    [0, 1, 2]
    [0, 2, 1]
    [1, 0, 2]
    [1, 2, 0]
    [2, 0, 1]
    [2, 1, 0]

    >>> permutation(seq) == permutation(*seq)
    True

    """
    # XXX
    # by using arbitrary argument lists,
    # you can use expressions both
    # permutation(('a','b','c'))    # -> single argument of tuple
    # and
    # permutation('a', 'b', 'c')    # -> three arguments of string

    if len(seq) == 1:
        seq = seq[0]

    # list method, 'pop', is used in perm_gen
    if not isinstance(seq, list):
        try:
            seq = list(seq)
        except TypeError, e:
            raise TypeError, "can't convert to list: %r"%(repr(seq))

    if not seq:
        #  seq = [], range(0), etc.

        # XXX
        # permutation([]) == [] or [[]]?

        # empty set
        return []
    elif len(seq) == 1:
        # seq = [1], ['a'], etc.
        return [seq]
    else:
        return list(perm_gen(seq))


def permhalf(seq):
    # permutation has a symmetric structure.
    # symmetric group <--> permutation

    # XXX
    # auxiliary function for perm_gen
    pop, insert, append = seq.pop, seq.insert, seq.append
    llen = len(seq)
    if llen <= 2:
        yield seq
    else:
        aRange = range(llen)
        v = pop()
        for p in permhalf(seq):
            for j in aRange:
                insert(j, v)
                yield seq
                del seq[j]
        append(v)

def half_perm_gen(seq):
    """generate half of all permutations from a given sequence.
    """
    ph = permhalf(seq)

    for h in ph:
        # XXX
        # don't forget to add [:] at yield statement.
        yield h[:]

def perm_gen(seq):
    """generate all permutations from a given sequence.
    """
    # largely owes the idea from comp.lang.python community.

    ph = permhalf(seq)

    for h in ph:
        p = h[:]

        # XXX
        # don't forget to add [:] at yield statement.
        yield p[:]
        p.reverse()
        yield p[:]

# NOTE
# next function is included in the distribution.
# see the following script.
# /src/Lib/test/test_generators.py
# the original name was 'gcomb'
def com_gen(seq, k):
    """Generate all combinations of k elements from the seq.
    """
    if k > len(seq):
        return
    if k == 0:
        yield []
    else:
        first, rest = seq[0], seq[1:]
        # A combination does or doesn't contain first.
        # If it does, the remainder is a k-1 comb of rest.
        for c in com_gen(rest, k-1):
            c.insert(0, first)
            yield c
        # If it doesn't contain first, it's a k comb of rest.
        for c in com_gen(rest, k):
            yield c

def catalan(num):
    """catalan(n) -> Return the Catalan number

    >>> for i in range(10):
    ...     print catalan(i)
    ...
    1
    1
    2
    5
    14
    42
    132
    429
    1430
    4862

    """
    # just for fun

    # C_n = (1 / (n+1)) * (2n, n)
    #
    #     = (1 / (n+1)) * (2n)! / (n!)^2
    #
    #     = (2n)! / ((n+1)! * n!)
    #
    #     = (2n * (2n-1) * ... * (n+1) / (n+1) !
    #
    #     = (2n * (2n-1) * ... * (n+2) / n !

    assert isinstance(num, (int, long)) and num >= 0, "must be a non-negative integer."
    if num in (0, 1):
        return 1

    return factorial(num+2, 2 * num) / factorial(num)

def catalan_generator():
    """catalan_generator() -> \
Return the Catalan number as a generator.

    >>> g = catalan_generator()
    >>> for i in range(10):
    ...     print g.next()
    ...
    1
    1
    2
    5
    14
    42
    132
    429
    1430
    4862
    """
    for i in itertools.count(0):
        yield catalan(i)

def k_composition(n, k):
    """k_composition(n, k) -> Return the number of k-composition of n.

    >>> n = 5
    >>> for i in range(n):
    ...     print k_composition(n, i+1)
    ...
    1
    4
    6
    4
    1

    """
    # Stanley EC1 p.14
    assert k-1 >= 0, "second argument must be non-negative"
    return combination(n-1, k-1)

def weak_k_composition(n, k):
    """weak_k_composition(n, k) -> Return the number of weak k-composition of n.

    >>> n = 5
    >>> for i in range(n):
    ...     print weak_k_composition(n, i+1)
    ...
    1
    6
    21
    56
    126
    
    """
    # Stanley EC1 p.15

    assert k-1 >= 0, "second argument must be non-negative"
    return combination(n + k -1, k-1)

@util.memoized
def signless_stirling1(n, m):
    """Return the signless Stirling number of the first kind

    >>> n = 5
    >>> for i in range(1, 6):
    ...     print signless_stirling1(n, i)
    ...
    24
    50
    35
    10
    1
    """
    # Stanley P.18

    if n == m == 0:
        return 1
    elif n <= 0 or m <= 0:
        return 0
    else:
        assert n > 0
        assert m > 0
        return (n-1) * signless_stirling1(n-1, m) + signless_stirling1(n-1, m-1)


@util.memoized
def stirling1(n, m):
    """Return the Stirling number of the first kind

    >>> n = 5
    >>> for i in range(1, 6):
    ...     print stirling1(n, i)
    ...
    24
    -50
    35
    -10
    1
    """
    # Stanley P.18

    sign = (n-m)%2 and -1 or 1
    return sign * signless_stirling1(n, m)

@util.memoized
def stirling2(n, m):
    """Return the Stirling number of the second kind
    
    >>> n = 5
    >>> for i in range(1,6):
    ...     print stirling2(n, i)
    ...
    1
    15
    25
    10
    1

    """
    # Stanley P. 33

    assert n >= 0
    assert m >= 0

    if n == m == 0:
        # By convention, we put S(0,0) = 1.
        return 1

    if n >= 1:
        if m > n:
            return 0
        elif m == 0:
            return 0
        elif m == 1:
            return 1

    return  m * stirling2(n-1, m) + stirling2(n-1, m-1)

def _test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    _test()
