# -*- coding: utf-8 -*-
# vi:ts=4:et
# $Id: util.py,v 1.6 2005/04/16 07:09:56 etale Exp $
# =====================================================
"""library for utility functions.
"""
__all__ = [
        'flatten_tableau',
        'int2str',
        'is_integer',
        'is_non_negative_integer',
        'memoized',
        'pairwise',
        'sequence_splitter',
        'trace',
        'transpose',
        'transpose_generator',
        ]

import itertools

def is_integer(n):
    """is_integer(n) -> Tests whether or not n is an integer.
    >>> is_integer(5)
    True
    >>> is_integer('a')
    False
    """
    return isinstance(n, (int, long))

def is_non_negative_integer(n):
    """is_non_negative_integer(n) -> Tests whether or not n is a non-negative integer.
    
    >>> is_non_negative_integer(5)
    True
    >>> is_non_negative_integer(0)
    True
    >>> is_non_negative_integer(-1)
    False
    >>> is_non_negative_integer([])
    False
    """
    return is_integer(n) and n >= 0

def int2str(seq, separator = " "):
    """convert a sequence of integers to a string
    
    >>> print int2str([3,1,2])
    3 1 2
    >>> print int2str([3,1,2], " * ")
    3 * 1 * 2
    """
    seq = itertools.imap(str, seq)

    return separator.join(seq)

def flatten_tableau(iterable, reverse=False):
    """This function is used to flatten a tableau
    >>> iterable = [[1,2],[3,4],[5]]
    >>> list(flatten_tableau(iterable))
    [1, 2, 3, 4, 5]
    """
    if reverse:
        iterable = reversed(iterable)
    for row in iterable:
        for col in row:
            yield col

def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ...
    >>> iterable = (0,1,2,3)
    >>> list(pairwise(iterable))
    [(0, 1), (1, 2), (2, 3)]
    """
    # This function is taken from an example in the library.
    # See : http://docs.python.org/lib/itertools-example.html

    a, b = itertools.tee(iterable)
    try:
        b.next()
    except StopIteration:
        pass
    return itertools.izip(a, b)

def transpose(matrix):
    """Transpose a matrix
    >>> matrix = [[1,0,1], [0,1,2],[3,4,5]]
    >>> transpose(matrix)
    [[1, 0, 3], [0, 1, 4], [1, 2, 5]]
    """
    return map(list, transpose_generator(matrix))

def transpose_generator(matrix):
    """Transpose a matrix(generator based)
    """
    return itertools.imap(None, *matrix)

def transpose_sequence(sequence):
    """Transpose a sequence
    """
    # TODO
    raise NotImplementedError

def transpose_sequence_generator(sequence):
    """Transpose a sequence(generator based)
    """
    # TODO
    raise NotImplementedError

def sequence_splitter(sequence):
    """
    split a sequence by a weakly increasing order.
    >>> seq = (1,2,2,1,2,1,1,1,2)
    >>> list(sequence_splitter(seq))
    [[1, 2, 2], [1, 2], [1, 1, 1, 2]]
    """

    prev = None
    tmp  = []
    for x in sequence:
        if x >= prev:
            tmp.append(x)
        else:
            yield tmp
            tmp = [x]
        prev = x
    else:
        if tmp:
            yield tmp

def trace(matrix):
    """Return the trace of a given matrix.
    >>> matrix = [[1,0,1], [0,1,2],[3,4,5]]
    >>> trace(matrix)
    7
    """
    return sum(matrix[i][i] for i in xrange(len(matrix)))

class memoized(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """
    # http://www.python.org/moin/PythonDecoratorLibrary
    def __init__(self, func):
        self.func = func
        self.cache = {}
    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            self.cache[args] = value = self.func(*args)
            return value
        except TypeError:
           # uncachable -- for instance, passing a list as an argument.
           # Better to not cache than to blow up entirely.
           return self.func(*args)
    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__


def _test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    _test()
